import asyncio
import logging

import httpx

from contact_magic.conf.settings import SETTINGS
from contact_magic.dict_utils import get_first_level_items, get_values_in_object_by_key
from contact_magic.logger import logger

logging.getLogger("httpx").setLevel(logging.WARNING)

BASE_URL = "https://app.copyfactory.io/api"

DEFAULT_TIMEOUT: float = 300.0


def _get_rate_limit_sleep_time(response):
    """Get rate limit window expiration time from response if the response
    status code is 429.
    """
    try:
        data = response.headers
        if "Retry-After" in data.keys():
            return int(data["Retry-After"])
    except (AttributeError, KeyError, ValueError):
        return 60


async def make_copyfactory_request(premise_id: int, variables: dict, max_retries=3):
    if not SETTINGS.COPYFACTORY_API_KEY:
        return None
    headers = {
        "Accept": "application/json",
        "Authorization": SETTINGS.COPYFACTORY_API_KEY,
    }
    variable_data = get_first_level_items(variables)
    data = {
        "premise_id": premise_id,
        "variables": variable_data,
    }
    try:
        async with httpx.AsyncClient(timeout=httpx.Timeout(DEFAULT_TIMEOUT)) as session:
            retries = 0
            while retries < max_retries:
                res = await session.request(
                    method="POST",
                    url=f"{BASE_URL}/v2/generate/",
                    headers=headers,
                    json=data,
                )
                # authorization error so can break
                if res.status_code == 403:
                    logger.warning("copyfactory_error", message=res.text)
                    break
                if res.status_code == 429:
                    await asyncio.sleep(_get_rate_limit_sleep_time(res))
                    retries += 1
                    continue

                if res.status_code == 200:
                    response_data = res.json()["data"]
                    if response_data["status"] == "success":
                        return response_data
                    if response_data["status"] == "error":
                        required_variables = response_data["meta_data"][
                            "sentence_premise"
                        ]["required_variables"]
                        for req_var in required_variables:
                            if item := list(
                                get_values_in_object_by_key(variables, req_var)
                            ):
                                if item[0]:
                                    data["variables"][req_var] = str(item[0])
                        if not data["variables"].get("company_organization_name"):
                            data["variables"]["company_organization_name"] = ""

                retries += 1
        return None
    except Exception:
        return None


def get_sequence_id_by_name(sequence_name: str, profile_id: int):
    headers = {
        "accept": "application/json",
        "Authorization": f"Bearer {SETTINGS.COPYFACTORY_API_KEY}",
        "Content-Type": "application/json",
    }
    data = {
        "sequence_name": sequence_name,
        "profile_id": profile_id,
    }
    res = httpx.get(
        url=f"{BASE_URL}/v3/sequences/", headers=headers, params=data
    ).json()
    if res:
        return res[0].get("id")
    res = httpx.post(url=f"{BASE_URL}/v3/sequences/", headers=headers, json=data).json()
    return res.get("id")


async def create_contact_in_copyfactory(
    sequence_id: str, mapping_id, email, sentence_fields: dict, custom_fields: dict
):
    headers = {
        "accept": "application/json",
        "Content-Type": "application/json",
        "Authorization": f"Bearer {SETTINGS.COPYFACTORY_API_KEY}",
    }
    json_data = {
        "email": email,
        "sequence_id": sequence_id,
        "custom_fields": get_first_level_items(custom_fields),
        "sentence_fields": get_first_level_items(sentence_fields),
        "mapping_id": mapping_id,
        "create_new_custom_fields": True,
        "create_new_sentence_fields": True,
    }
    async with httpx.AsyncClient(timeout=httpx.Timeout(DEFAULT_TIMEOUT)) as session:
        res = await session.post(
            url="https://app.copyfactory.io/api/v3/contacts/",
            headers=headers,
            json=json_data,
        )
        return res.json()



"""
Callbacks only accept the row that was finished and the current campaign.

Callbacks must return the DataRow.

"""
from contact_magic import DataRow, PersonalizationSettings
from contact_magic.dict_utils import get_first_level_items
from contact_magic.integrations.copyfactory import create_contact_in_copyfactory


async def add_as_contact_to_copyfactory(
    data_row: DataRow, campaign: PersonalizationSettings
) -> DataRow:
    """
    Agency callback to automatically add new
    processed contacts to Copyfactory to trigger
    a sync with a CRM or sales automation platform.
    """
    email_col_value = data_row.get_values_by_key_name(
        "Email"
    ) or data_row.get_values_by_key_name("email")
    if not email_col_value:
        return data_row
    # These are generated columns we don't want to push to CF.
    custom_cols_exclude = campaign.get_datapoint_column_names + [
        "domain_to_check",
        "location_search_from",
        "search_query",
        "email",
        "is_approved",
        "Email",
    ]
    await create_contact_in_copyfactory(
        sequence_id=campaign.campaign_meta_data.get("sequence_id"),
        mapping_id=campaign.campaign_meta_data.get("mapping_id"),
        email=email_col_value[0],
        sentence_fields={
            k.replace("__", "_"): data_row.data.get(k)
            for k in campaign.get_datapoint_column_names
            if not k.startswith("source__")
        },
        custom_fields={
            f"cus_{k}": v
            for k, v in get_first_level_items(data_row.data).items()
            if k not in custom_cols_exclude and not k.startswith("technology__")
        },
    )
    return data_row



import asyncio
import logging

import httpx

from contact_magic.conf.settings import SETTINGS
from contact_magic.dict_utils import get_first_level_items
from contact_magic.logger import logger
from contact_magic.utils import fix_website

logging.getLogger("httpx").setLevel(logging.WARNING)

base_url = "https://salesscrapers.dev/api/"


async def make_sales_scraper_request(endpoint, data: dict, max_retries=3):
    if not SETTINGS.SALES_SCRAPERS_API_KEY:
        return None
    headers = {
        "Accept": "application/json",
        "X-API-Key": SETTINGS.SALES_SCRAPERS_API_KEY,
    }

    url = f"{base_url}{endpoint}"
    try:
        retries = 0
        params = get_first_level_items(data)
        while retries < max_retries:
            res = await poll_until_job_complete(url, headers, params)
            if not res:
                return {}
            # authorization error so can break
            if res.get("job_status_code") == 401:
                logger.warning("sales_scraper_error", message=res.get("job_data"))
                break
            if res.get("job_status_code") == 400:
                logger.warning("proxy_error", message=res.get("job_data"))
            if res.get("job_status_code") == 200:
                return res.get("job_data")
            if res.get("job_status_code") == 422:
                validation_errors = res.get("job_data").get("detail")
                for error in validation_errors:
                    if error.get("type") == "value_error.url.scheme":
                        error_key = error.get("loc")[1]
                        if error_key in data:
                            params[error_key] = fix_website(data[error_key])
            retries += 1
        return None
    except Exception:
        return None


async def poll_until_job_complete(url, headers, params) -> dict:
    new_job = await create_new_job(url, headers, params)
    # raise auth error from SS or validation error
    if new_job.status_code == 422:
        return {"job_status_code": new_job.status_code, "job_data": new_job.json()}
    if new_job.status_code != 200:
        return {"job_status_code": new_job.status_code, "job_data": new_job.text}
    new_job_id = new_job.json().get("job_id")
    job_is_pending = True
    count_sleep = 0
    max_wait_rotations = 150
    while job_is_pending:
        if count_sleep >= max_wait_rotations:
            return {}
        job_status = await get_job_status(new_job_id, headers)
        if job_status.get("job_status") == "SUCCESS":
            return job_status
        await asyncio.sleep(15)
        count_sleep += 1
    return {}


async def create_new_job(url, headers, params):
    """
    Create new job in Salescrapers. This returns
    {
      "job_id": "f41ecfb9-78b2-4214-b503-444634550ec4",
      "job_status": "SUCCESS",
      "job_status_code": 200,
      "job_data": {}
      }
    :param url:
    :param headers:
    :param params:
    :return:
    """
    async with httpx.AsyncClient(timeout=httpx.Timeout(300.0)) as session:
        res = await session.request(
            method="get", url=url, headers=headers, params=params
        )
        return res


async def get_job_status(job_id, headers) -> dict:
    """
    Get job status in Salescrapers. This returns
    {
      "job_id": "f41ecfb9-78b2-4214-b503-444634550ec4",
      "job_status": "SUCCESS",
      "job_status_code": 200,
      "job_data": {}
      }
    """
    async with httpx.AsyncClient(timeout=httpx.Timeout(300.0)) as session:
        res = await session.request(
            method="get", url=f"{base_url}job-status/{job_id}", headers=headers
        )
        return res.json()


async def get_custom_scrapers():
    async with httpx.AsyncClient(timeout=httpx.Timeout(300.0)) as session:
        headers = {
            "Accept": "application/json",
            "X-API-Key": SETTINGS.SALES_SCRAPERS_API_KEY,
        }
        res = await session.request(
            method="get", url=base_url + "custom-scrapers", headers=headers
        )
        return res.json()



ALL_CATEGORIES = [
    "CMS",
    "Message boards",
    "Database managers",
    "Documentation",
    "Widgets",
    "Ecommerce",
    "Photo galleries",
    "Wikis",
    "Hosting panels",
    "Analytics",
    "Blogs",
    "JavaScript frameworks",
    "Issue trackers",
    "Video players",
    "Comment systems",
    "Security",
    "Font scripts",
    "Web frameworks",
    "Miscellaneous",
    "Editors",
    "LMS",
    "Web servers",
    "Caching",
    "Rich text editors",
    "JavaScript graphics",
    "Mobile frameworks",
    "Programming languages",
    "Operating systems",
    "Search engines",
    "Webmail",
    "CDN",
    "Marketing automation",
    "Web server extensions",
    "Databases",
    "Maps",
    "Advertising",
    "Network devices",
    "Media servers",
    "Webcams",
    "Payment processors",
    "Tag managers",
    "CI",
    "Control systems",
    "Remote access",
    "Development",
    "Network storage",
    "Feed readers",
    "DMS",
    "Page builders",
    "Live chat",
    "CRM",
    "SEO",
    "Accounting",
    "Cryptominers",
    "Static site generator",
    "User onboarding",
    "JavaScript libraries",
    "Containers",
    "PaaS",
    "IaaS",
    "Reverse proxies",
    "Load balancers",
    "UI frameworks",
    "Cookie compliance",
    "Accessibility",
    "Authentication",
    "SSL/TLS certificate authorities",
    "Affiliate programs",
    "Appointment scheduling",
    "Surveys",
    "A/B Testing",
    "Email",
    "Personalisation",
    "Retargeting",
    "RUM",
    "Geolocation",
    "WordPress themes",
    "Shopify themes",
    "Drupal themes",
    "Browser fingerprinting",
    "Loyalty & rewards",
    "Feature management",
    "Segmentation",
    "WordPress plugins",
    "Hosting",
    "Translation",
    "Reviews",
    "Buy now pay later",
    "Performance",
    "Reservations & delivery",
    "Referral marketing",
    "Digital asset management",
    "Content curation",
    "Customer data platform",
    "Cart abandonment",
    "Shipping carriers",
    "Shopify apps",
    "Recruitment & staffing",
    "Returns",
    "Livestreaming",
    "Ticket booking",
    "Augmented reality",
    "Cross border ecommerce",
    "Fulfilment",
    "Ecommerce frontends",
    "Domain parking",
    "Form builders",
    "Fundraising & donations",
]

DEFAULT_TRACKABLE_CATEGORIES = [
    "CMS",
    "Ecommerce",
    "Analytics",
    "LMS",
    "Marketing automation",
    "Advertising",
    "Payment processors",
    "Tag managers",
    "DMS",
    "Page builders",
    "Live chat",
    "CRM",
    "SEO",
    "Accounting",
    "User onboarding",
    "Affiliate programs",
    "Appointment scheduling",
    "Surveys",
    "A/B Testing",
    "Email",
    "Personalisation",
    "Retargeting",
    "Loyalty & rewards",
    "Reviews",
    "Buy now pay later",
    "Reservations & delivery",
    "Referral marketing",
    "Cart abandonment",
    "Shipping carriers",
    "Shopify apps",
    "Recruitment & staffing",
    "Returns",
    "Ticket booking",
    "Augmented reality",
    "Cross border ecommerce",
    "Fulfilment",
    "Form builders",
    "Fundraising & donations",
]
